/*
 *    Rosegarden MIDI Sequencer
 *
 *    File:           Sequence_LINUX.c
 *
 *    Description:    Sequencer playback functions.  Provides a generic
 *                    core to which the various hardware specific things
 *                    can be bolted, switches permitting.  Looking for
 *                    an elegant way of accomplishing this dynamically,
 *                    or indeed at all.
 *
 *    Author:         AJG; SGI dmedia MIDI library support by Archer Sully,
 *                    archer@elysium.esd.sgi.com, integrated into this
 *                    file by cc.
 *
 *    History:
 *
 *    Update  Date            Programmer      Comments
 *    ======  ====            ==========      ========
 *
 *    001     29/03/94        AJG             File Created.
 *    002     09/08/94        JPff            Error check in Midi_SeqOutInit
 *    003     17/02/95        cc              Added #ifdefs for HAVE_Z8530
 *    004     20.2.96         rwb             Changes to make sequencing
 *                                            support more portable.
 *    005     24.4.96         rwb             Completed initial Playback
 *                                            changes under Linux. 
 *    006     30.5.96         rwb             Midi Out and Synth Support.
 *
 */


#undef POSIX_PLEASE
#undef _POSIX_SOURCE

/* include our linux soundcard header */
#include <sys/soundcard.h>
#include <sys/ioctl.h>
#include <errno.h>

/* now the Rosegarden stuff */
#include "Globals.h"
#include "Consts.h"

#include <MidiXInclude.h>
#include <MidiFile.h>
#include <MidiErrorHandler.h>
#include <MidiBHeap.h>
#include <MidiTrack.h>

#include "Sequence.h"

#include <stdio.h>
#include <sys/types.h>
#include <fcntl.h>

#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include <Debug.h>

/* ++++ Structures ++++ */
typedef union _DeviceInformation
{
    struct midi_info   Midi;
    struct synth_info  Synth;
} DeviceInformation;

typedef struct _DeviceListElement
{
    ListElement       Scooter;
    int               Type;
    DeviceInformation Port;
} DeviceListElement, *DeviceList;


/* ++++ Predeclarations ++++ */
void seqbuf_dump();
void Midi_SeqPlayFileCB();
int Midi_SeqIntCB();
void Midi_SeqAllNotesOff();
void Midi_SeqWriteMidiEvent();
DeviceList Midi_NewDeviceList();
void Midi_DeviceCloseCB();
void Midi_GetDeviceInfo();
void Midi_SeqWriteSynthEvent();

char *MidiPortName;  /* we use this for the device, not the MIDI port */
char *MidiDevice;
char *SynthDevice;
int MidiDev;
int SynthDev;
int SynthEnable = 0; /* for playback only, toggles between MIDI and synth */
int seqfd;


#define MidiPlaybackEventsMask MidiSoundEventsMask | MidiSetTempoEventMask

#define MIDI_EVENT_HEAP_SIZE 1024

EventList  *FilteredTracks;
BinaryHeap  EventHeap;
Boolean     Playing;
Widget      StopButtonDlg;
Widget      DeviceDlg;
int         MidiPort;

SEQ_DEFINEBUF(100); /* then define a buffer for the API */

/*
 * Function:     Midi_DeviceQuery
 *
 * Description:  Opens device and queries for all Midi ports
 *               and available synths.  If everything
 *
 */
DeviceList Midi_DeviceQuery(char *Device)
{
    int midi_devs = 0;
    int synth_devs = 0;
    DeviceInformation TempDevice;
    DeviceList Devices = NULL,
               CurrentDev = NULL;
    int i;

BEGIN("Midi_QueryDevice");


    /* open device */
    if ( Midi_OpenDevice(O_WRONLY, Device) != True )
        return NULL;

    /* call to see if there are any synths on-line */
    ioctl(seqfd, SNDCTL_SEQ_NRSYNTHS, &synth_devs);

    /* see if we have any MIDI devices available */
    ioctl(seqfd, SNDCTL_SEQ_NRMIDIS, &midi_devs);

    if ( midi_devs == synth_devs == 0 )
    {
/*
        Error(NON_FATAL_REPORT_TO_MSGBOX,"No Configurable Devices on Port\n");
*/
        close(seqfd);
        return NULL;
    }

    /* construct the devices return list */
    for ( i = 0; i < synth_devs; i++ )
    {
        /* initialise the structure */
        TempDevice.Synth.device = i;

        ioctl(seqfd, SNDCTL_SYNTH_INFO, &(TempDevice.Synth));

        if ( Devices == NULL )
        {
            Devices = (DeviceList)Midi_NewDeviceList(TempDevice, SYNTH_DEVICE);
        }
        else
        {
             /* append */
            CurrentDev = (DeviceList)Midi_NewDeviceList(TempDevice,
                                                            SYNTH_DEVICE);
            LIST_Insert ( (List) Devices, (List) CurrentDev );
        }
    }

    for ( i = 0; i < midi_devs; i++ )
    {
        TempDevice.Midi.device = i;
        ioctl(seqfd, SNDCTL_MIDI_INFO, &(TempDevice.Midi));

        if ( Devices == NULL )
        {
            Devices = (DeviceList)Midi_NewDeviceList(TempDevice, MIDI_DEVICE);
        }
        else
        {
            /* again append */
            CurrentDev = (DeviceList)Midi_NewDeviceList(TempDevice,
                                                                MIDI_DEVICE);
            LIST_Insert ( (List) Devices, (List) CurrentDev );
        }
    }

    /* finished gathering information - close the port down */
    close(seqfd);

    /* return a stack of pertinent data to the calling function */
    return Devices;

    END;
}



/*
 * Function:     Midi_SeqStopPlayingCB
 *
 * Description:  Destory the transient widget, clear down all
 *               the output channels and then close the port.
 *
 */
void Midi_SeqStopPlayingCB(Widget w, XtPointer a, XtPointer b)
{
BEGIN("Midi_SeqStopPlayingCB");

    /* clear all MIDI channels */
    Midi_SeqAllNotesOff();

    /* free the midi port */
    close(seqfd);

    Playing = False;

END;
}


/*
 * Function:     Midi_SeqPlayFileCB
 *
 * Description:  Once a tree.
 *
 */
void
Midi_SeqPlayFileCB(Widget w, XtPointer a, XtPointer b)
{
    int       i;
    Widget    StopButton;
    XPoint    op;
    EventList Temp;

BEGIN("Midi_SeqPlayFileCB");

    /* open the device, which port we're using we don't
       case about just yet */
    if ( Midi_OpenDevice(O_WRONLY, MidiPortName ) != True )
        return;

    /* right, create a binary heap, spacey likes these */
    EventHeap = CreateBHeap(MIDI_EVENT_HEAP_SIZE, Midi_EventListTimeLessp);

    /* do a Xt malloc to the size of the event list */
    FilteredTracks = (EventList *)XtMalloc(MIDIHeaderBuffer.NumTracks *
                                                           sizeof(EventList));

    /*************************************************************/
    /* Produce filtered versions of the tracks in the MIDI file. */
    /* Filter out any non-soundable events - these can simply be */
    /* ignored during playback.                                  */
    /*************************************************************/

    for (i = 0; i < MIDIHeaderBuffer.NumTracks; ++i)
    {
        Temp = Midi_TrackFilterByEvent(MIDITracks[i], MidiPlaybackEventsMask);

        FilteredTracks[i] = Midi_TrackConvertToTwoPointRepresentation(Temp);

        Midi_TrackDelete(Temp); /* so we've got a temp list of two points */

        if (FilteredTracks[i]) BHeapInsert(EventHeap, FilteredTracks[i]);
    }

    Playing = True;   /* start the loop and away we go */

    /* process any pending X events */
    while(XtAppPending(appContext))
    {
        XtAppProcessEvent(appContext, XtIMAll);
    }

    Midi_SeqIntCB();  /* check this at some point for success or not */

    /* close the output and the transient if we haven't already */
    if ( Playing == True )
        Midi_SeqStopPlayingCB(NULL, NULL, NULL);

END;
}



/*
 * Function:     Midi_SeqIntCB
 *
 * Description:  Play the events.
 *
 */
int
Midi_SeqIntCB()
{
    EventList         NextEvent;
    unsigned int      TempoValue;
    unsigned int      LastTime = 0;
    float             Deltatime = 0.0;
    float             PlayTime = 0.0;
    float             TimeInc;

    BEGIN("Midi_SeqIntCB");

    /* reset the midi */
    ioctl(seqfd, SNDCTL_SEQ_RESET, 0);

    /* NextEvent->Event.DeltaTime is giving us the absolute timing
       from the beginning of the track - we want to convert this
       by the division and the other timing information into an
       absolute wait for all events to be queued.  We queue the
       vents up and DUMP them out when we need them played.
       USS Lite takes care of this for us.  Thanks Hannu.  */

    /* start the timer */
    SEQ_START_TIMER();

    while ( ( NextEvent = (EventList)Value(EventHeap, Root) ) &&
                                ( Playing == True ) )
    {
        /* do we need to queue an event ? */

        if ( ( NextEvent->Event.DeltaTime ) > LastTime )
        {
            Deltatime = ( ( NextEvent->Event.DeltaTime ) - LastTime ) * TimeInc;
            PlayTime += Deltatime;

#ifdef DEBUG
            fprintf(stdout,"Division = %f\n",Deltatime);
#endif

            SEQ_WAIT_TIME( (int) PlayTime );
#ifdef DEBUG
            fprintf(stdout,"wait for %f ticks (%ld)\n", PlayTime,
                                                NextEvent->Event.DeltaTime); 
#endif
/*  synchonisation in here 
            SEQ_DUMPBUF();
*/

            LastTime = NextEvent->Event.DeltaTime;
        }

        /* beware disguised macros */

        if (NextEvent->Event.EventCode == MIDI_FILE_META_EVENT)
        {
            if (NextEvent->Event.EventData.MetaEvent.MetaEventCode ==
                                                          MIDI_SET_TEMPO)
            {

              /* getting the tempo FINE */

              TempoValue = (NextEvent->Event.EventData.MetaEvent.Bytes
                 << 16) | (*(&NextEvent->Event.EventData.MetaEvent.Bytes
                 + 1) << 8) | *(&NextEvent->Event.EventData.MetaEvent.Bytes
                 + 2);

#ifdef DEBUG
              fprintf(stdout,"Division set to %d\n", MIDIHeaderBuffer.
                                                     Timing.Division);
#endif
              TimeInc = TempoValue / ( 10000.0 * MIDIHeaderBuffer.
                                                        Timing.Division );
            }
        }
        else
        {
            /* write the event out to the desired device */
            if ( SynthEnable )
                Midi_SeqWriteSynthEvent(NextEvent);
            else
                Midi_SeqWriteMidiEvent(NextEvent);
        }

        /* shuffle around and get the next event */
        ExtractMin(EventHeap);
        if (Next(NextEvent))
        {
            BHeapInsert(EventHeap, Next(NextEvent));
        }

        if (!HeapSize(EventHeap))
        {
            /* shut down and this sets Playing to False */
            Midi_SeqStopPlayingCB(NULL, NULL, NULL);
        }
        while (XtAppPending(appContext)) XtAppProcessEvent(appContext, XtIMAll);
    }

    RETURN_BOOL(True);  /* return an OK */
}



/*
 * Function:     seqbuf_dump
 *
 * Description:  Write the queued events out of the selected port.
 *
 */
void
seqbuf_dump()
{
    if ( _seqbufptr )
        if ( write ( seqfd, _seqbuf, _seqbufptr ) == -1 )
        {
            /* write has failed */
            perror ("write /dev/sequencer");
        }
    _seqbufptr = 0;

    /* synchonisation call - we wait for the alloted period of time */
    ioctl(seqfd, SNDCTL_SEQ_SYNC);
}        



/*
 * Function:     Midi_SeqWriteMidiEvent
 *
 * Description:  Write the events onto the output buffer according to type
 *
 */
void Midi_SeqWriteMidiEvent(EventList OutEvent)
{
    byte        StatusByte;
    static byte LastStatusByte = 0;

    BEGIN("Midi_SeqWriteMidiEvent");

#ifdef DEBUG
    fprintf(stdout,"Writing out Event Data %x\n",OutEvent->
                                      Event.EventData.NoteOn.Note);
#endif
    
    StatusByte = OutEvent->Event.EventCode;

    if (StatusByte != LastStatusByte)
    {
        LastStatusByte = StatusByte;
        SEQ_MIDIOUT(MidiDev,StatusByte);
    }

    if (MessageType(StatusByte) == MIDI_CTRL_CHANGE ||
        MessageType(StatusByte) == MIDI_PROG_CHANGE)
    {
        SEQ_MIDIOUT(MidiDev,(byte)OutEvent->Event.EventData.
                                                   ProgramChange.Program);
    }
    else
    {
        /* out with an event */
        SEQ_MIDIOUT(MidiDev,(byte)OutEvent->Event.EventData.NoteOn.Note);
        SEQ_MIDIOUT(MidiDev,(byte)OutEvent->Event.EventData.NoteOn.Velocity);
    }

END;
}


/*
 * Function:     Midi_SeqWriteSynthEvent
 *
 * Description:  Kagagoogoo
 *
 */
void
Midi_SeqWriteSynthEvent(EventList OutEvent)
{
    byte        StatusByte;
    static byte LastStatusByte = 0;
    byte Note = (byte)OutEvent->Event.EventData.NoteOn.Note;
    byte Vely = (byte)OutEvent->Event.EventData.NoteOn.Velocity;
    byte Channel;

BEGIN("Midi_SeqWriteMidiEvent");

    StatusByte = OutEvent->Event.EventCode;

    if (StatusByte != LastStatusByte)
    {
        LastStatusByte = StatusByte;
        SEQ_MIDIOUT(MidiDev,StatusByte);
    }

    Channel = StatusByte&0xf;
    switch(MessageType(StatusByte))
    {
        case MIDI_KEY_PRESSURE:
            SEQ_KEY_PRESSURE(SynthDev, Channel, Note, Vely);
            break;

        case MIDI_NOTEON:
            SEQ_START_NOTE(SynthDev, Channel, Note, Vely);
            break;

        case MIDI_NOTEOFF:
            SEQ_STOP_NOTE(SynthDev, Channel, Note, Vely);
            break;

        case MIDI_CTL_CHANGE:
             SEQ_CONTROL(SynthDev, Channel, Note, Vely);
             break;

        case MIDI_CHN_PRESSURE:
             SEQ_CHN_PRESSURE(SynthDev, Channel, Note);
             break;

        case MIDI_PITCH_BEND:
              SEQ_BENDER(SynthDev, Channel, Note);

        case MIDI_PGM_CHANGE:
 
        default:
            break;

    }

END;
}

/*
 * Function:     Midi_SeqAllNotesOff
 *
 * Description:  Clear down all channels on the specified device.
 *
 */
void Midi_SeqAllNotesOff(void)
{
    int i;
    byte CmdBuffer[3];

    BEGIN("Midi_SeqAllNotesOff");

    CmdBuffer[1] = 123;
    CmdBuffer[2] = 0;

    for(i = 0; i < 16; ++i)
    {
        /* note that CreateMessageByte is another well hidden macro */
        CmdBuffer[0] = CreateMessageByte(MIDI_CTRL_CHANGE, i);
        SEQ_MIDIOUT(MidiDev,(byte)CmdBuffer[0]);
        SEQ_MIDIOUT(MidiDev,(byte)CmdBuffer[1]);
        SEQ_MIDIOUT(MidiDev,(byte)CmdBuffer[2]);
    }

    SEQ_DUMPBUF();
    END;
}


/*
 * Function:     Midi_NewDeviceList
 *
 * Description:  Dbly Linked Lists constructor for the
 *               returning of synth and midi device info.
 *
 */
DeviceList
Midi_NewDeviceList( DeviceInformation NewDevice, int Type )
{
    DeviceList newlist;

    BEGIN("MidiNewDeviceList");

    newlist = (DeviceList) NewList (sizeof(DeviceListElement));
    newlist->Port = NewDevice;
    newlist->Type = Type;

    return;
    END;
}



void
Midi_PlaybackDlg( DeviceList Devices )
{
    Widget DevicePane;
    Widget DeviceTopBox;
    Widget DeviceSelectedLabel;
    Widget DeviceSelectedField;
    Widget DeviceLabel;
    Widget DeviceForm;
    Widget DeviceBottomBox;
    Widget DeviceOK;
    Widget DeviceClose;
    Widget DevicePlay;
    Widget DeviceStop;
    Widget DeviceHelp = NULL;
    XPoint          op;

    BEGIN("Midi_PlaybackDlg");

    DeviceDlg = XtCreatePopupShell("Devices",transientShellWidgetClass,
                                                   topLevel, NULL, 0);

    DevicePane = YCreateWidget("Device Pane", panedWidgetClass, DeviceDlg);

    DeviceTopBox = YCreateShadedWidget("Device Title Box", boxWidgetClass,
                                          DevicePane, MediumShade);

    DeviceLabel = YCreateLabel( "MIDI Playback", DeviceTopBox );

    DeviceForm = YCreateShadedWidget ("Device Form", formWidgetClass, 
                                 DevicePane,    LightShade);

    /* select message according to output device */
    if ( SynthEnable )
    {
        DeviceSelectedLabel = YCreateLabel("Synthesiser Playback Device:",
                                                                DeviceForm);
        DeviceSelectedField = YCreateLabel(SynthDevice, DeviceForm);
    }
    else
    {
        DeviceSelectedLabel = YCreateLabel("Midi Playback Device :",
                                                                DeviceForm);
        DeviceSelectedField = YCreateLabel(MidiDevice, DeviceForm);
    }
        

    DeviceBottomBox = YCreateShadedWidget ("Device Bottom Box", boxWidgetClass,
                                 DevicePane, MediumShade);

    DevicePlay = YCreateCommand("Play", DeviceBottomBox);
    DeviceStop = YCreateCommand("Stop", DeviceBottomBox);
    DeviceClose = YCreateCommand("Close", DeviceBottomBox);

    YSetValue(XtParent(DeviceStop),XtNfromHoriz,XtParent(DevicePlay));
    YSetValue(XtParent(DeviceClose),XtNfromHoriz,XtParent(DeviceStop));
    YSetValue(XtParent(DeviceSelectedField),XtNfromHoriz,
                                         XtParent(DeviceSelectedLabel));

    XtAddCallback(DevicePlay, XtNcallback, Midi_SeqPlayFileCB, NULL);
    XtAddCallback(DeviceStop, XtNcallback, Midi_SeqStopPlayingCB, NULL);
    XtAddCallback(DeviceClose, XtNcallback, Midi_DeviceCloseCB, NULL);

    op = YPlacePopupAndWarp(DeviceDlg, XtGrabNonexclusive,
                                DeviceClose, DeviceStop);

    END;
}

void
Midi_DeviceCloseCB(Widget w, XtPointer a, XtPointer b)
{
    BEGIN("Midi_DeviceCancelCB");

    /* close down playing */
    Midi_SeqStopPlayingCB(NULL, NULL, NULL);

    /* and clear up */
    YPopdown(DeviceDlg);
    XtDestroyWidget(DeviceDlg);
    END;
}


int
Midi_OpenDevice(int Sense, char *Device)
{
    char        ErrBuff[128];

    BEGIN("Midi_OpenDevice");

    if ( ( seqfd = open ( Device, Sense, 0 ) ) == -1 )
    {
        switch (errno)
        {
            case ENOENT:
                sprintf(ErrBuff, "\"%s\" : Device File Missing\n", Device);
                break;

            case ENODEV:
                sprintf(ErrBuff, "\"%s\" : Device Driver Not Installed\n",
                                                     Device);
                break;

            case ENXIO:
                sprintf(ErrBuff, "\"%s\" : No Hardware detected\n", Device);
                break;

            case EBUSY:
                sprintf(ErrBuff, "Device \"%s\" is busy\n", Device);
                break;

            default:
                sprintf(ErrBuff, "\"%s\" : Cannot Open Device\n", Device);
                break;
        }
        Error(NON_FATAL_REPORT_TO_MSGBOX, ErrBuff);
        return False;
    }

    return True;

    END;
}



void
Midi_GetDeviceInfo(void)
{
    DeviceList Extract = NULL, Current;
    DeviceList Devices;
    int i;

BEGIN("Midi_GetDeviceInfo");

    /* initialise all device parameters as "missing" */
    MidiDev = -1;
    SynthDev = -1;
    MidiDevice = strdup("No Midi Device");
    SynthDevice = strdup("No Synthesiser");

    if ( ( Devices = Midi_DeviceQuery(MidiPortName) ) == 0 )
    {
        /* no devices found - return devices as unavailable */
        return;
    }

    /* return synth name and device number plus
       midi port name and device number */

    Current = (DeviceList) LIST_First( (List) Devices );

    while(Current)
    {
        if ( Current->Type == MIDI_DEVICE )
        {
            MidiDev = Current->Port.Midi.device;
            MidiDevice = strdup(Current->Port.Midi.name);
        }
        if ( Current->Type == SYNTH_DEVICE )
        {
            SynthDev = Current->Port.Synth.device;
            SynthDevice = strdup(Current->Port.Synth.name);
        }
        Current = (DeviceList) Next(Current);
    }

END;
}
